#include <includes.h>
#include <console_drv.h>
#include "intr.h"
#include "md.h"
#include "opt_string.h"



md_spin_lock_t temp_console_lock = MD_SPIN_UNLOCKED;

/*
 * If the irq handlers are using printk, 
 * We need to disable interrupt to prevent deadlock
 */
/*
#define LOCK_CONSOLE()
#define UNLOCK_CONSOLE()
*/

#define LOCK_CONSOLE()\
	local_irq_disable();\
	md_spin_lock( & temp_console_lock );
    
#define UNLOCK_CONSOLE()\
	md_spin_unlock( & temp_console_lock );\
	local_irq_enable();


#define MD_CURSOR_COLS (80u)
#define MD_CURSOR_ROWS (25u)

#define MD_MAX_CURSOR_OFFSET ( MD_CURSOR_COLS * MD_CURSOR_ROWS )

#define MD_U8P_VRAM ((u8*)(0xB8000))
#define MD_PUT_CHAR( offset, ch )		*( MD_U8P_VRAM + (offset << 1) ) = ch

#define CURSOR_QUERY_PORT	(0x3D4)
#define CURSOR_DATA_PORT	(0x3D5)
#define CURSOR_OFFSET_MSB	(14)
#define CURSOR_OFFSET_LSB	(15)

OS_INLINE u16 md_get_cursor_offset()
{
	u16 offset;
	md_out( CURSOR_QUERY_PORT, CURSOR_OFFSET_MSB );
	offset = md_in( CURSOR_DATA_PORT ) << 8;
	
	md_out( CURSOR_QUERY_PORT, CURSOR_OFFSET_LSB );
	offset |= md_in( CURSOR_DATA_PORT );
	
	return offset;
}

OS_INLINE void md_set_cursor_offset(u16 offset)
{
	md_out( CURSOR_QUERY_PORT, CURSOR_OFFSET_MSB );
	md_out( CURSOR_DATA_PORT, (unsigned char) ( offset >> 8 ) );
	
	md_out( CURSOR_QUERY_PORT, CURSOR_OFFSET_LSB );
	md_out( CURSOR_DATA_PORT, (unsigned char) ( offset ) );
}


typedef struct cursor_system_t 
{
	u32 x;
	u32 y;
	u16 offset;
} cursor_system_t;

cursor_system_t cursor;

u16 line_offsets[ MD_CURSOR_ROWS ];

#define U16_CURSOR_OFFSET_Y( y ) ( line_offsets[y] )

#define CURSOR_OFFSET_LAST_LINE     MD_CURSOR_COLS * ( MD_CURSOR_ROWS - 1 )

#define CURSOR_X( offset ) ( offset % MD_CURSOR_COLS )
#define CURSOR_Y( offset ) ( offset / MD_CURSOR_COLS )


#if defined(DEBUG)	
void DEBUG_CURSOR_POS( u8 ch )
{
	if ( (cursor.x == 0) && (cursor.y == 0) )
	{
		MD_PUT_CHAR( cursor.offset, ch );
		for(;;);
	}
}
#else
#define DEBUG_CURSOR_POS(a)
#endif


void con_scroll_up()
{
	u8 * vidmem_last_line = 
		MD_U8P_VRAM + MD_CURSOR_COLS * ( MD_CURSOR_ROWS - 1 ) * 2 ;
	u32 i;
	
	
	memcpy(		MD_U8P_VRAM, 
				MD_U8P_VRAM + MD_CURSOR_COLS * 2, 
				(size_t) ( MD_CURSOR_COLS * ( MD_CURSOR_ROWS - 1 ) * 2 ) );
				
	for (i=0; i < MD_CURSOR_COLS; i ++ ) {
		*vidmem_last_line++ = 0;
		*vidmem_last_line++ = 0xF;
	}
}


void init_console()
{
	u32 y;
	
	LOCK_CONSOLE();
	
	cursor.offset	= md_get_cursor_offset();
	cursor.x		= CURSOR_X( cursor.offset );
	cursor.y		= CURSOR_Y( cursor.offset );
	
	/*
	 * cursor offset lookup table for performance
	 */
	for ( y=0; y < MD_CURSOR_ROWS; y++ )
	{
		line_offsets[ y ] = (u16) (y * MD_CURSOR_COLS ) ;
	}

	UNLOCK_CONSOLE();
}

OS_INLINE u32 con_get_cursor_x() 
{
	return (cursor.offset % MD_CURSOR_COLS);
}


OS_INLINE u32 con_get_cursor_y() 
{
	return (cursor.offset / MD_CURSOR_COLS);
}


void con_set_cursor_pos( const u32 x, const u32 y )
{
	cursor.offset = U16_CURSOR_OFFSET_Y( y ) + x ;
	cursor.x = x;
	cursor.y = y;
	md_set_cursor_offset( cursor.offset );
}



void put_char_ ( const u8 ch )
{
	
	MD_PUT_CHAR( cursor.offset, ch );

	if ( cursor.x < MD_CURSOR_COLS - 1 )
	{
		cursor.x ++;
		cursor.offset ++;
	}
	else
	{
		if ( cursor.y < MD_CURSOR_ROWS - 1 )
		{
			cursor.y ++ ;
			cursor.x = 0;
			
			cursor.offset ++;
		}
		else
		{
			con_scroll_up();
			
			cursor.offset = CURSOR_OFFSET_LAST_LINE;
			cursor.x = 0;
			cursor.y = MD_CURSOR_ROWS - 1;
			
		}
	}
	
}

void put_char( const u8 ch )
{
	if ( ch == '\n' )
	{
		if ( cursor.y < MD_CURSOR_ROWS -1 )
		{
			cursor.y ++ ;
			cursor.x = 0;
		}
		else
		{
			cursor.x = 0;
			con_scroll_up();
		}
		
		cursor.offset = U16_CURSOR_OFFSET_Y( cursor.y );
	}
	else
	{
		put_char_( ch );
	}
}

/*
 * Text Printing API
*/
void con_putc( const u8 ch )
{
	put_char( ch );
	
	md_set_cursor_offset( cursor.offset );
}

void con_put_U32( const u8 * message, const u32 data )
{
	const u8 * digit = "0123456789ABCDEF";
	s32 nibble;


	con_puts("[0x");

	for ( nibble = 7; nibble >=0; nibble -- )
	{
		con_putc( digit[ ( data >> ( nibble * 4 ) ) & 0xF ] );
	}

	con_puts("] ");

	con_puts(message);
	
	con_putc('\n');

}


void con_puts( const u8 * str )
{
	LOCK_CONSOLE();
	
	while( *str )
	{
		put_char( *str++ );
	}
	
	md_set_cursor_offset( cursor.offset );
	
	UNLOCK_CONSOLE();
}


void con_clear()
{
	u8 *vidmem = MD_U8P_VRAM;
	const u32 size  = MD_CURSOR_COLS * MD_CURSOR_ROWS ;
	u32 loop;

	LOCK_CONSOLE();
	
	for (loop=0; loop<size; loop ++ ) {
		*vidmem++ = 0; 
		*vidmem++ = 0xF;
	}
	
	con_set_cursor_pos( 0, 0 );
	
	UNLOCK_CONSOLE();
}
