#include <includes.h>
#include <console.h>
#include "../mc146818rtc.h"
#include "../md.h"
#include "../atomic.h"
#include "../intr.h"
#include "smp.h"
#include "apic.h"
#include "apic_def.h"
#include "mpparse.h"


static atomic_t init_deasserted;


static s32 wakeup_secondary_cpu(u32 phys_apicid, u32 start_eip)
{
	unsigned long send_status = 0, accept_status = 0;
	int maxlvt, timeout, num_starts, j;

	/*
	 * Be paranoid about clearing APIC errors.
	 */
	if (APIC_INTEGRATED(apic_version[phys_apicid])) {
		apic_read_around(APIC_SPIV);
		apic_write(APIC_ESR, 0);
		apic_read(APIC_ESR);
	}

	Dprintk("Asserting INIT.\n");

	/*
	 * Turn INIT on target chip
	 */
	apic_write_around(APIC_ICR2, SET_APIC_DEST_FIELD(phys_apicid));

	/*
	 * Send IPI
	 */
	apic_write_around(APIC_ICR, 
		 APIC_INT_LEVELTRIG | APIC_INT_ASSERT	| APIC_DM_INIT);

	Dprintk("Waiting for send to finish...\n");
	timeout = 0;
	do {
		Dprintk("+");
		udelay(100);
		send_status = apic_read(APIC_ICR) & APIC_ICR_BUSY;
	} while (send_status && (timeout++ < 1000));

	mdelay(10);

	Dprintk("Deasserting INIT.\n");

	/* Target chip */
	apic_write_around(APIC_ICR2, SET_APIC_DEST_FIELD(phys_apicid));

	/* Send IPI */
	apic_write_around(APIC_ICR, APIC_INT_LEVELTRIG | APIC_DM_INIT);

	Dprintk("Waiting for send to finish...\n");
	timeout = 0;
	do {
		Dprintk("+");
		udelay(100);
		send_status = apic_read(APIC_ICR) & APIC_ICR_BUSY;
	} while (send_status && (timeout++ < 1000));

	atomic_set(&init_deasserted, 1);

	/*
	 * Should we send STARTUP IPIs ?
	 *
	 * Determine this based on the APIC version.
	 * If we don't have an integrated APIC, don't send the STARTUP IPIs.
	 */
	if (APIC_INTEGRATED(apic_version[phys_apicid]))
		num_starts = 2;
	else
		num_starts = 0;

	/*
	 * Run STARTUP IPI loop.
	 */
	Dprintk("#startup loops: %d.\n", num_starts);

	maxlvt = get_maxlvt();

	for (j = 1; j <= num_starts; j++) {
		Dprintk("Sending STARTUP #%d.\n",j);
		apic_read_around(APIC_SPIV);
		apic_write(APIC_ESR, 0);
		apic_read(APIC_ESR);
		Dprintk("After apic_write.\n");
		/*
		 * STARTUP IPI
		 */

		/* Target chip */
		apic_write_around(APIC_ICR2, SET_APIC_DEST_FIELD(phys_apicid));


		/* Boot on the stack */
		/* Kick the second */
		apic_write_around(APIC_ICR, APIC_DM_STARTUP
									| (start_eip >> 12));

		/*
		 * Give the other CPU some time to accept the IPI.
		 */
		udelay(300);
		

		Dprintk("Startup point 1.\n");

		Dprintk("Waiting for send to finish...\n");

		timeout = 0;
		do {
			Dprintk("+");
			udelay(100);
			send_status = apic_read(APIC_ICR) & APIC_ICR_BUSY;
		} while (send_status && (timeout++ < 1000));

		/*
		 * Give the other CPU some time to accept the IPI.
		 */
		udelay(200);
		/*
		 * Due to the Pentium erratum 3AP.
		 */
		if (maxlvt > 3) {
			apic_read_around(APIC_SPIV);
			apic_write(APIC_ESR, 0);
		}
		accept_status = (apic_read(APIC_ESR) & 0xEF);
		if (send_status || accept_status)
			break;
	}
	Dprintk("After Startup.\n");

	if (send_status)
		printk("APIC never delivered???\n");
	if (accept_status)
		printk("APIC delivery error (%lx).\n", accept_status);

	return (send_status | accept_status);
}

static s32 do_boot_cpu(u32 phys_apicid, u32 start_eip)
/*
 * NOTE - on most systems this is a PHYSICAL apic ID, but on multiquad
 * (ie clustered apic addressing mode), this is a LOGICAL apic ID.
 * Returns zero if CPU booted OK, else error code from wakeup_secondary_cpu.
 */
{
	s32 boot_error;
//	unsigned short nmi_high = 0, nmi_low = 0;


	atomic_set(&init_deasserted, 0);

	Dprintk("Setting warm reset code and vector.\n");

//	store_NMI_vector(&nmi_high, &nmi_low);


	smpboot_setup_warm_reset_vector(start_eip);

	
	/*
	 * Starting actual IPI sequence...
	 */
	boot_error = wakeup_secondary_cpu(phys_apicid, start_eip);

	return boot_error;
}

static void startup_APs()
{
	// for calibrating 
	u32 apicid = 0u;
	u32 kicked = 1u;
	
	Dprintk(KERN_INFO "Booting Application Processors. (Total Count = %d)\n", num_processors);
	
	for (apicid = 0; kicked < num_processors && apicid < MAX_APICS; apicid++) {
		
		/*
		 * Don't even attempt to start the boot CPU!
		 */
		if ( apicid == boot_cpu_apicid )
		{
			Dprintk("[ APICID = %d ] Bypassing BSP.\n", apicid);
			continue;
		}
		
		if (apicid == BAD_APICID)
		{
			Dprintk("[ APICID = %d ] Bypassing bad APICID.\n", apicid);
			continue;
		}

//		if (!check_apicid_present(bit))
//			continue;
//		if (max_cpus <= cpucount+1)
//			continue;

		printk(KERN_INFO "Trying to startup CPU with APIC #%d ...", apicid );

		if ( do_boot_cpu(apicid, AP_STARTUP_EIP) != 0 )
		{
			printk(KERN_INFO "Failed \nCPU APIC #%d not responding - cannot use it.\n",
								apicid);
		}
		else
		{
			++kicked;
			printk("OK\n");
		}
		
		Dprintk("kicked:%d, num_processors:%d, apicid:%d, MAX_APICS:%d\n", 
				kicked, num_processors, apicid, MAX_APICS );
	}

	printk(KERN_INFO "Finished booting Application Processors.\n");
	printk(KERN_INFO "Currently %d CPUs are actively working.\n", cpu_count );

	/*
	 * Cleanup possible dangling ends...
	 */
	smpboot_restore_warm_reset_vector();
	
	
	/*
	 * If Hyper-Threading is avaialble, construct cpu_sibling_map[], so
	 * that we can tell the sibling CPU efficiently.
	 */
/*	 
	if (cpu_has_ht && smp_num_siblings > 1) {
		for (cpu = 0; cpu < NR_CPUS; cpu++)
			cpu_sibling_map[cpu] = NO_PROC_ID;
		
		for (cpu = 0; cpu < NR_CPUS; cpu++) {
			int 	i;
			if (!cpu_isset(cpu, cpu_callout_map))
				continue;

			for (i = 0; i < NR_CPUS; i++) {
				if (i == cpu || !cpu_isset(i, cpu_callout_map))
					continue;
				if (phys_proc_id[cpu] == phys_proc_id[i]) {
					cpu_sibling_map[cpu] = i;
					printk("cpu_sibling_map[%d] = %d\n", cpu, cpu_sibling_map[cpu]);
					break;
				}
			}
			if (cpu_sibling_map[cpu] == NO_PROC_ID) {
				smp_num_siblings = 1;
				printk(KERN_WARNING "WARNING: No sibling found for CPU %d.\n", cpu);
			}
		}
	}
	*/
}








void init_system()
{
	if ( ! have_cpuid_p()) {
		init_console();
		printk(KERN_ERR " Unsupported CPU. CPUID instruction is not available.\n" );
		halt();
	}
	
	cpu_info_t * cpu = & mach_.cpu[ thr_cpu_id() ];
	
	identify_cpu( cpu );
}

OS_INLINE u32 is_bsp() 
{
/*	
	thread_info_t * ti = current_thread_info();
	init_console();
	printk("ti => %x \n", ti );
	printk("*(u32*)ti => %d \n", *(u32*)ti );
	printk("ti->cpu => %d \n", ti->cpu );
*/	
	return thr_cpu_id() == 0 ;
}

void init_every_cpu()
{
	
	/* setup IDT and enable interrupts */
	init_idt();

	cpu_info_t * cpu = & mach_.cpu[ thr_cpu_id() ];
	map_lapic( cpu );
	
	init_apic_intr();
}

/*
 * Only boot processor should run this.
 */
//#include "../profiler.h"

void init_BSP()
{
	init_console();

/*	
	PROF_INIT();
	con_puts("12345");
	PROF_OFF();
	PROF_PRINT();
	for(;;);
*/
	
	/* Remap IRQ from 0x20 to 0x2F and disable all interrupts by default*/
	remap_8259A_irq( 0x20, 0x28 );
	
	/* enable A20 line */
	init_cpu_A20();
	
	
	find_smp_config();
	get_smp_config();

	init_every_cpu();
	
	setup_8254_timer();
	
	local_irq_enable();
	
	verify_local_APIC();

	connect_bsp_APIC();

	setup_local_APIC();

/*
	if (nmi_watchdog == NMI_LOCAL_APIC)
		check_nmi_watchdog();
#ifdef CONFIG_X86_IO_APIC
	if (smp_found_config)
		if (!skip_ioapic_setup && nr_ioapics)
			setup_IO_APIC();
#endif
*/
	setup_boot_APIC_clock();
	
	/* wakeup application processors */
	startup_APs();
	
	enable_APIC_timer();
	
	printk(KERN_INFO "BSP APIC Timer has just started.\n" );
}




void init_APs()
{


	printk(KERN_INFO "Initializing CPU %d...\n", thr_cpu_id() );
		
	/*
	 * If waken up by an INIT in an 82489DX configuration
	 * we may get here before an INIT-deassert IPI reaches
	 * our local APIC.  We have to wait for the IPI or we'll
	 * lock up on an APIC access.
	 */
	 
	 
    while (!atomic_read( & init_deasserted));
	

	init_every_cpu();
	
//	clear_local_APIC();
	
//	verify_local_APIC();
	
	setup_local_APIC();
	
	setup_secondary_APIC_clock();
/*		
		if (nmi_watchdog == NMI_IO_APIC) {
			disable_8259A_irq(0);
			enable_NMI_through_LVT0(NULL);
			enable_8259A_irq(0);
		}
*/		         
	enable_APIC_timer();

	local_irq_enable();

	printk(KERN_INFO "APIC Timer of CPU %d has just started...\n", thr_cpu_id() );
}




